#!/usr/local/bin/python2.6

#
# Copyright (C) 2009, Oy L M Ericsson Ab, NomadicLab <pubsub@nomadiclab.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2 as
# published by the Free Software Foundation.
#
# Alternatively, this software may be distributed under the terms of the BSD
# license.
#
# See LICENSE and COPYING for more details.
#

"""PSIRP Local Area Intradomain Rendezvous Daemon."""


from psirp.libpsirp import *
from psirp.ps_debug import *
import psirp.libpsirp_py # XXX

import select, signal, re, time, traceback
#import pdb # debugger, add breakpoint: pdb.set_trace()


ROOT_SCOPE_SID        = buffer(PSIRP_ID_LEN*'\x00')
DEFAULT_IPC_SID       = atoid("::bb00")
DEFAULT_IPC_LOCAL_RID = atoid("::bb01")
DEFAULT_IPC_NET_RID   = atoid("::bb02")
EMPTY_RID             = buffer(PSIRP_ID_LEN*'\x00')
EMPTY_FID             = buffer(PSIRP_FID_LEN*'\x00')
#FULL_FID              = buffer(PSIRP_FID_LEN*'\xff')

# Everything else except "::" (Scope 0) and DEFAULT_IPC_SID (::bb00)
# are network scopes:
DEFAULT_LOCAL_SID_PATTERN = "(?!^(::|::bb00)$)"


class Subscriber(object):
    
    class Subs(object):
        
        def __init__(self, parent):
            self.parent = parent # Subscriber
            self.net_subs = SIdRIdDict()
            self.non_subs = SIdRIdDict(count=True)
        
        # Pseudo-publication for getting events from the subs file
        class SubsPub(Publication):
            
            def __init__(self, subs_file, parent):
                super(type(self), self).__init__(None) # XXX
                
                self.subs_file = subs_file
                self.pep = libpsirp_py.ps_event_page()
                self.parent = parent
            
            def __del__(self):
                try:
                    self.subs_file.close()
                except:
                    pass
            
            @property
            def fd(self):
                return self.subs_file.fileno()
            
            def read(self):
                return libpsirp_py.psirp_py_read_event_page(self.fd, self.pep)
            
            def handle_event(self, event, pub):
                debug_enter()
                
                subs = pub.read() # This "publication" actually wraps a file
                debug_print(PSIRP_DBG_GARB, "%d subscriptions" % len(subs))
                
                for sid, rid in subs:
                    # Make copies, beacause data in the underlying
                    # page can change
                    sid, rid = copybuf(sid), copybuf(rid)
                    sid_str = idtoa(sid)
                    
                    debug_print(PSIRP_DBG_GARB,
                                "Subscription to SId/RId: %s/%s"
                                % (sid_str, idtoa(rid)))

                    # Is this a subscription done by ourselves?
                    if self.parent.non_subs.contains(sid, rid):
                        self.parent.non_subs.remove(sid, rid)
                        debug_print(PSIRP_DBG_GARB,
                                    "Subscription ignored "
                                    "(self-inflicted)")
                        continue # for sid, rid in subs
                    
#                    # Has this already been subscribed?
#                    elif self.parent.net_subs.contains(sid, rid):
#                        debug_print(PSIRP_DBG_INFO,
#                                    "Subscription ignored "
#                                    "(already subscribed)")
#                        # Should we support re-subscribing?
#
#                        continue # for sid, rid in subs
                    
                    # Is this a non-network scope?
                    elif re.match(self.parent.parent.sid_pattern,
                                  sid_str) is None:
                        debug_print(PSIRP_DBG_GARB,
                                    "Subscription ignored "
                                    "(%s is not a network scope)" % sid_str)
                        continue # for sid, rid in subs
                    
                    # Is this publication already available locally?
                    # XXXXX
                    else:
                        try:
                            self.parent.non_subs.add(sid, rid)
                            pub = subscribe(sid, rid)
                            
                            debug_print(PSIRP_DBG_GARB,
                                        "Subscription ignored "
                                        "(already stored locally)")
                            continue # for sid, rid in subs
                        except NotFoundError, e:
                            # Publication not available locally
                            pass
                    
                    if self.parent.parent.net.metadata.contains(sid, rid):
                        self.parent.net_subs.add(sid, rid)
                        debug_print(PSIRP_DBG_INFO,
                                    "Already have metadata, "
                                    "subscribing to data, RId=%s"
                                    % idtoa(rid))
                        self.parent.parent.net.publish_ipc_md_ext(
                            libpsirp_py.PSIRP_HDR_RZV_SUBSCRIBE_DATACHUNK,
                            self.parent.parent.net.metadata[sid][rid],
                            EMPTY_FID,
                            False)
                    else:
                        self.parent.net_subs.add(sid, rid)
                        debug_print(
                            PSIRP_DBG_INFO,
                            "Subscription stored, SId/RId=%s"
                            % idstoa(sid, rid))
                        
                        # First we need to subscribe to the metadata
                        debug_print(PSIRP_DBG_INFO,
                                    "Subscribing to metadata "
                                    "from the RZV node, RId=%s"
                                    % idtoa(rid))
                        self.parent.parent.net.publish_ipc_md(
                            libpsirp_py.PSIRP_HDR_RZV_SUBSCRIBE_METADATA,
                            Metadata(sid, rid))

        def init(self, pskq):
            debug_enter()
            
            FILENAME = "/pubsub/subs"
            subs_file = open(FILENAME)
            
            subs_pub = self.SubsPub(subs_file, self)
            subs_pub_fd = subs_pub.fd
            subs_pub_fflags = \
                NOTE_PUBLISH | select.KQ_NOTE_WRITE | select.KQ_NOTE_ATTRIB
            pskq.custom_command(subs_pub.fd, fflags=subs_pub_fflags)
            pskq.pubs[subs_pub_fd] = subs_pub # XXX
            
    
    class Net(object):

        def __init__(self, parent, ipc_sid, ipc_local_rid, ipc_net_rid):
            self.parent = parent # Subscriber
            self.ipc_sid = ipc_sid
            self.ipc_local_rid = ipc_local_rid
            self.ipc_net_rid = ipc_net_rid
            
            self.metadata = SIdRIdDict()
        
        def init(self, pskq):
            debug_enter()
            
            # Subscribe to network-triggered events (inbound in this context)
            try:
                self.parent.subs.non_subs.add(self.ipc_sid,
                                              self.ipc_net_rid)
                ipc_pub = subscribe(self.ipc_sid, self.ipc_net_rid)
            except NotFoundError, e:
                debug_print(PSIRP_DBG_INFO,
                            "IPC publication does not yet exist, creating it")
                
                createpub(self.ipc_sid, self.ipc_net_rid)
                
                self.parent.subs.non_subs.add(self.ipc_sid,
                                              self.ipc_net_rid)
                ipc_pub = subscribe_sync(self.ipc_sid, self.ipc_net_rid, 1.0)
            
            ipc_pub.handle_event = self.get_event_handler()
            pskq.register(ipc_pub, True, True); del ipc_pub
        
        def get_event_handler(self):
            def handle_event(event, ipc_pub):
                debug_enter()
                
                metadata = libpsirp_py.psirp_py_pub_hdrs_metadata(ipc_pub.pubp)
                
                sid  = copybuf(metadata.getsid())
                rid  = copybuf(metadata.getrid())
                vrid  = copybuf(metadata.getvrid())
                # Collected FId to the pub'r or sub'r:
                fid  = copybuf(metadata.getfid())
                len  = metadata.len
                type = metadata.rzvhdr_type
                
                rid_str = idtoa(rid)
                
#                debug_print(PSIRP_DBG_INFO,
#                            "Received metadata:\n"
#                            "         SId/RId: %s\n"
#                            "         vRId:    %s\n"
#                            "         FId:     %s\n"
#                            "         len:     %d\n"
#                            "         type:    0x%02x"
#                            % (idstoa(sid, rid),
#                               idtoa(vrid),
#                               idtoa(fid),
#                               len,
#                               type))
                
                if type == libpsirp_py.PSIRP_HDR_RZV_PUBLISH_METADATA:
                    # Somebody in the network published something,
                    # or we have subscribed to a publication.
                    debug_print(PSIRP_DBG_INFO,
                                "Received 'publish metadata' "
                                "from the blackboard")
                    
                    timestamp = time.time()
                    
                    # Check if publication exists locally
                    try:
                        self.parent.subs.non_subs.add(sid, rid)
                        pub = subscribe(sid, rid)
                        
                        # We already have this publication.
                        # Perhaps somebody else published a new
                        # version or something, but in this
                        # implementation we currently just drop
                        # the metadata.
                        return
                    except NotFoundError, e:
                        # OK, we can proceed since we don't seem to have it yet.
                        pass
                    
                    # Add to metadata map
                    md = Metadata(sid, rid, vrid, fid, len, timestamp)
                    self.metadata.add(sid, rid, md)
                    debug_print(PSIRP_DBG_INFO,
                                "Metadata stored, RId=%s" % rid_str);
                    
                    # If we have this in our subscription map, we have
                    # previously subscribed to it and want to get the data.
                    if self.parent.subs.net_subs.contains(sid, rid):
                        debug_print(PSIRP_DBG_INFO,
                                    "Subscribe to data, RId=%s" % rid_str);
                        
                        self.publish_ipc_md_ext(
                            libpsirp_py.PSIRP_HDR_RZV_SUBSCRIBE_DATACHUNK,
                            md,
                            EMPTY_FID,
                            False)
                
                elif type == libpsirp_py.PSIRP_HDR_RZV_SUBSCRIBE_METADATA:
                    # Somebody in the network subscribed to metadata
                    debug_print(PSIRP_DBG_INFO,
                                "Recevied 'subscribe metadata' "
                                "from the blackboard")
                    
                    if self.metadata.contains(sid, rid):
                        # We have the corresponding metadata, so let's
                        # publish it
                        md = self.metadata[sid][rid]
                        
#                        debug_print(PSIRP_DBG_INFO,
#                                    "Publishing (external) metadata "
#                                    "to the subscriber, "
#                                    "RId=%s" % rid_str);
                        
                        md_out = Metadata(sid, rid, md.vrid, fid, md.len)
                        self.publish_ipc_md(
                            libpsirp_py.PSIRP_HDR_RZV_PUBLISH_METADATA,
                            md_out)
                    else:
                        # Check if we have the publication locally
                        try:
                            self.parent.subs.non_subs.add(sid, rid)
                            spub = subscribe(sid, rid)
                            if spub.type != PS_PUB_DATA:
                                debug_print(
                                    PSIRP_DBG_WARN,
                                    "Subscription to non-data publication")
                                return
                            spub_vrid = spub.vrid
                        except NotFoundError, e:
                            debug_print(
                                PSIRP_DBG_WARN,
                                "Subscription to unknown publication")
                            # XXX: In order to support sub-before-pub
                            # cases we should also store subscriptions
                            # that came from the net, especially if we
                            # are a local RZV node...
                            return
                        
#                        debug_print(PSIRP_DBG_INFO,
#                                    "Publishing (local) metadata "
#                                    "to the subscriber, "
#                                    "RId=%s" % rid_str);
                        
                        md_out = Metadata(sid, rid, spub_vrid, fid, spub.len)
                        self.publish_ipc_md(
                            libpsirp_py.PSIRP_HDR_RZV_PUBLISH_METADATA,
                            md_out)
                
                elif type == libpsirp_py.PSIRP_HDR_RZV_SUBSCRIBE_DATACHUNK:
                    # Somebody in the network subscribed to data
                    debug_print(PSIRP_DBG_INFO,
                                "Received 'subscribe data' from the blackboard")
                    
                    if self.metadata.contains(sid, rid):
                        # We have the corresponding metadata, but
                        # presumably not the data. We pass on this
                        # subscription to the publisher.
                        debug_print(PSIRP_DBG_INFO,
                                    "Relay subscription to data chunks")
                        md = self.metadata[sid][rid]
                        self.publish_ipc_md_ext(
                            libpsirp_py.PSIRP_HDR_RZV_SUBSCRIBE_DATACHUNK,
                            md,
                            fid,
                            True)
                    else:
                        # Check if we have this publication locally.
                        # If we have it, send it out.
                        try:
                            # RId in scope?
                            self.parent.subs.non_subs.add(sid, rid)
                            pub = subscribe(sid, rid)
                            if (vrid != EMPTY_RID):
                                # Version-RId in publication?
                                #
                                # XXX: We don't know if this version was
                                #      published in the given scope;
                                #      currently that just cannot be verified.
                                self.parent.subs.non_subs.add(rid, vrid)
                                vpub = subscribe(rid, pub.vrid)
                                pub_vrid = vpub.vrid
                                pub_len  = vpub.len
                            else:
                                # Just get the latest version
                                pub_vrid = pub.vrid
                                pub_len  = pub.len
                            
                            md = Metadata(sid, rid, pub_vrid, fid, pub_len)
#                            debug_print(PSIRP_DBG_INFO,
#                                        "Publish data chunks:\n"
#                                        "         SId/RId: %s\n"
#                                        "         vRId:    %s\n"
#                                        "         FId:     %s\n"
#                                        "         len:     %d"
#                                        % (idstoa(sid, rid),
#                                           idtoa(pub_vrid),
#                                           idtoa(fid),
#                                           pub_len))
                            
                            # netiod also subscribes to the version...
                            self.parent.subs.non_subs.add(rid, vrid)
                            
                            self.publish_ipc_md(
                                libpsirp_py.PSIRP_HDR_RZV_PUBLISH_DATACHUNK,
                                md)
                        except NotFoundError, e:
                            debug_print(
                                PSIRP_DBG_INFO,
                                "Subscription to unknown publication")
            
                elif type == libpsirp_py.PSIRP_HDR_RZV_PUBLISH_DATACHUNK:
                    debug_print(PSIRP_DBG_WARN,
                                "Received 'publish datachunk' "
                                "from the blackboard "
                                "-- unexpectedly and unrequestedly!")
                
                else:
                    debug_print(PSIRP_DBG_WARN,
                                "Unknown IPC publication type: %d" % type)
                return
            return handle_event
        
        def publish_ipc_md(self, type, md):
            debug_enter()
            
            pub = Publication.from_pubp(
                    libpsirp_py.psirp_py_create_md_pub(None))
            metadata = libpsirp_py.psirp_py_pub_hdrs_metadata(pub.pubp)
            buf = pub.buffer
            
            metadata.rzvhdr_type = type
            metadata.hdr_type = libpsirp_py.PSIRP_HDR_MD
            metadata.setsid(md.sid)
            metadata.setrid(md.rid)
            metadata.setvrid(md.vrid)
            metadata.setfid(md.fid)
            metadata.len = md.len

            debug_print(PSIRP_DBG_INFO,
                        "Publishing metadata on the blackboard:\n%s"
                        % metadata)
            
            pub.publish(self.ipc_sid, self.ipc_local_rid)
            
            debug_return()
        
        def publish_ipc_md_ext(self, type, md, extra_fid, relay):
            debug_enter()
            
            pub = Publication.from_pubp(
                    libpsirp_py.psirp_py_create_md_ext_pub(None))
            metadata = libpsirp_py.psirp_py_pub_hdrs_metadata_ext(pub.pubp)
            buf = pub.buffer
            
            metadata.md_hdr.rzvhdr_type = type
            metadata.md_hdr.hdr_type = libpsirp_py.PSIRP_HDR_MD
            metadata.setmdsid(md.sid)
            metadata.setmdrid(md.rid)
            metadata.setmdvrid(md.vrid)
            metadata.setmdfid(md.fid)
            metadata.md_hdr.len = md.len
            
            metadata.setextfid(extra_fid)
            metadata.md_ext.relay = relay

            debug_print(PSIRP_DBG_INFO,
                        "Publishing metadata "
                        "on the blackboard:\n%s"
                        % metadata)
            
            pub.publish(self.ipc_sid, self.ipc_local_rid)
            
            debug_return()
        
    
    class Scope(object):
        
        def __init__(self, parent, parent_sid, sid):
            self.parent = parent
            self.parent_sid = parent_sid
            self.sid = sid
            
            self.rid_count = 0
        
        def init(self, pskq, is_net_scope=None):
            debug_enter()
            
            # Subscribe to scope
            debug_print(PSIRP_DBG_GARB, "Subscribe to scope")
            
            self.sid_str = idtoa(self.sid)
            if is_net_scope is None:
                self.is_net_scope = (re.match(self.parent.sid_pattern,
                                              self.sid_str) is not None) # XXX
            else:
                self.is_net_scope = is_net_scope
            
            self.parent.subs.non_subs.add(self.sid, self.sid)
            scope_pub = subscribe(self.sid, self.sid)
            
            debug_print(PSIRP_DBG_GARB,
                        "Read RIds in scope with SId: %s"
                        % idtoa(self.parent_sid))
            scope_data = scope_pub.get_rids()
            
            # Read and process initial scope contents
            for rid in scope_data:
                self.init_pub(pskq, copybuf(rid))
            
            self.rid_count = len(scope_data)
            debug_print(PSIRP_DBG_GARB,
                        "Scope initialized, %d RIds" % self.rid_count)
        
        def init_pub(self, pskq, rid):
            sid_str = idtoa(self.sid)
            rid_str = idtoa(rid)
            
            # Subscribe to this publication
            debug_print(PSIRP_DBG_GARB,
                        "Subscribe to SId/RId: %s/%s" % (sid_str, rid_str))
            self.parent.subs.non_subs.add(self.sid, rid)
            pub = subscribe(self.sid, rid)
            pub_len = pub.len
            
            pub.timestamp = time.time()
            
            vrid = copybuf(pub.vrid)
            vrid_str = idtoa(vrid)
            
            # Register to its events
            if pub.type == PS_PUB_SCOPE:
                is_net_scope = (re.match(self.parent.sid_pattern,
                                         rid_str) is not None) # XXX
                if rid == ROOT_SCOPE_SID or is_net_scope:
                    if rid == self.sid:
                        debug_print(PSIRP_DBG_INFO,
                                    "Registering to scope events, "
                                    "SId=%s" % rid_str)
                        pub.handle_event = self.get_scope_event_handler()
                        pskq.register(pub, True, True); del pub
                    else:
                        debug_print(PSIRP_DBG_GARB,
                                             "Initializing subscope, "
                                             "SId=%s" % rid_str)
                        subscope = self.parent.Scope(self.parent,
                                                     self.sid, rid)
                        subscope.init(pskq, is_net_scope)
                else:
                    debug_print(PSIRP_DBG_INFO,
                                "SId=%s is not a network scope" % rid_str)
            else:
                debug_print(PSIRP_DBG_INFO,
                            "Registering to publication events, "
                            "RId=%s" % rid_str)
                # XXX: This could be re-publication in another scope,
                #      but right now we always treat it as a new
                #      publication.
                pub.handle_event = self.get_pub_event_handler()
                pskq.register(pub, True, True); del pub
                
                if self.is_net_scope:
                    if self.parent.net.metadata.contains(self.sid, rid):
                        # Remove metadata
                        self.parent.net.metadata.remove(self.sid, rid)
                    
                    if self.parent.subs.net_subs.contains(self.sid, rid):
                        # This publication had been subscribed previously.
                        # Remove the pending subscription.
                        self.parent.subs.net_subs.remove(self.sid, rid)
                    else:
                        # Issue IPC command to get new metadata published
                        # into the network
#                        debug_print(PSIRP_DBG_INFO,
#                                    "Publishing metadata to the RZV node:\n"
#                                    "         SId/RId: %s/%s\n"
#                                    "         vRId:    %s\n"
#                                    "         len:     %d"
#                                    % (sid_str, rid_str, vrid_str,
#                                       pub_len))
                        md = Metadata(self.sid, rid, vrid, EMPTY_FID, pub_len)
                        self.parent.net.publish_ipc_md(
                            libpsirp_py.PSIRP_HDR_RZV_PUBLISH_METADATA, md)
                else:
                    debug_print(PSIRP_DBG_INFO,
                                "SId=%s is not a network scope"
                                % self.sid_str)
        
        def get_scope_event_handler(self):
            def handle_event(event, scope_pub):
                debug_enter()
                
                if event.fflags & NOTE_PUBLISH:
                    # Something was published in the scope
                    # XXX: We assume that items can only be added,
                    #      and not removed.
                    old_rid_count = self.rid_count
                    new_rids = scope_pub.get_rids(old_rid_count)
                    new_rid_count = old_rid_count + len(new_rids)
                    debug_print(PSIRP_DBG_GARB,
                                "Old RId count: %d, new RId count: %d"
                                % (old_rid_count, new_rid_count))
                    
                    if new_rid_count > old_rid_count:
                        for rid in new_rids:
                            debug_print(PSIRP_DBG_INFO,
                                        "New scope entry: SId/RId=%s"
                                        % idstoa(self.sid, rid))
                            self.init_pub(self.parent.pskq, copybuf(rid))
                        
                        self.rid_count = new_rid_count
                else:
                    debug_print(PSIRP_DBG_WARN, "Scope subscribed?")
                
                debug_return()
            
            return handle_event
        
        def get_pub_event_handler(self):
            def handle_event(event, pub):
                debug_enter()
                
                if event.fflags & NOTE_PUBLISH:
                    sid  = copybuf(pub.sid)
                    rid  = copybuf(pub.rid)
                    vrid = copybuf(pub.vrid)
                    
                    pub.timestamp = time.time()
                
                    debug_print(PSIRP_DBG_INFO,
                                "Publication update, SId/RId=%s"
                                % idstoa(sid, rid))
                    
                    # Issue IPC command to get new metadata published
                    # into the network
                    if self.is_net_scope:
                        md = Metadata(sid,
                                      rid,
                                      vrid,
                                      EMPTY_FID,
                                      pub.len)
                        self.parent.net.publish_ipc_md(
                            libpsirp_py.PSIRP_HDR_RZV_PUBLISH_METADATA, md)
                    else:
                        debug_print(PSIRP_DBG_INFO,
                                    "SId=%s is not a network scope"
                                    % self.sid_str)
                
                debug_return()
            
            return handle_event
    
    def __init__(self, sid_pattern, ipc_sid, ipc_local_rid, ipc_net_rid):
        self.sid_pattern = re.compile(sid_pattern)
        
        self.pskq = PubSubKQueue()
        self.subs = self.Subs(self)
        self.net = self.Net(self, ipc_sid, ipc_local_rid, ipc_net_rid)
        self.root_scope = self.Scope(self, ROOT_SCOPE_SID, ROOT_SCOPE_SID)
    
    def laird_listen(self):
        debug_enter()
        
        # Register to events and initialize handlers
        self.subs.init(self.pskq)
        self.net.init(self.pskq)
        self.root_scope.init(self.pskq)
        
        # Listen
        debug_print(PSIRP_DBG_INFO, "Listening to events")
        try:
            while True:
                publ = self.pskq.listen()
                for ev, pub in publ:
                    if not isinstance(pub, Publication):
                        debug_print(PSIRP_DBG_WARN,
                                    "Not a publication: %r, %r" % (ev, pub))
                        continue
                    pub.handle_event(ev, pub)
        except KeyboardInterrupt:
            debug_print(PSIRP_DBG_INFO, "Rendezvous listener interrupted")
        except:
            debug_print(PSIRP_DBG_ERR,
                        "Rendezvous listener:\n%s"
                        % traceback.format_exc())
        
        # Cleanup
        debug_print(PSIRP_DBG_GARB, "Cleaning up")
        self.pskq.close()
        
        debug_print(PSIRP_DBG_GARB, "Done")

    
def createpub(sid, rid):
    pub = create(1)
    pub.publish(sid, rid)
    return pub


def copybuf(buf):
    return buffer(buf[:])


class SIdRIdDict(dict):
    # {SId<buffer>: {RId<buffer>: counter<int>|item<object>, ...}, ...}
    
    def __init__(self, count=False):
        self.count = count
    
    def add(self, sid, rid, item=None):
        if sid not in self:
            self[sid] = {}
        if self.count and rid in self[sid]:
            self[sid][rid] += 1
        else:
            if item is None:
                item = 1
            self[sid][rid] = item
    
    def remove(self, sid, rid):
        if sid in self and rid in self[sid]:
            if self.count and self[sid][rid] > 1:
                self[sid][rid] -= 1
            else:
                del self[sid][rid]
            if len(self[sid]) == 0:
                del self[sid]
    
    def contains(self, sid, rid):
        return (self[sid][rid] if sid in self and rid in self[sid] else 0)


class Metadata(object):
    def __init__(self, sid, rid, vrid=EMPTY_RID, fid=EMPTY_FID, len=0,
                 timestamp=0):
        self.sid  = sid
        self.rid  = rid
        self.vrid = vrid
        self.fid  = fid
        self.len  = len
        self.timestamp = timestamp



def main():
    import getopt
        
    try:
        opts, args = getopt.gnu_getopt(sys.argv[1:], "s:r:q:t:ch", [])
        #print(opts, args)
    except getopt.GetoptError, err:
        print(str(err))
        sys.exit(1)
    
    debug_level = 0
    debug_colors = False
    sid_pattern = DEFAULT_LOCAL_SID_PATTERN
    ipc_sid = DEFAULT_IPC_SID
    ipc_local_rid = DEFAULT_IPC_LOCAL_RID
    ipc_net_rid = DEFAULT_IPC_NET_RID
    
    for o, a in opts:
        if o == "-h":
            print("%s %s" \
                      % (str(__file__).split("/")[-1],
                         "[-s sid_pattern -r ipc_local_rid -q ipc_net_rid "
                         "-t debug_level -c -h]"))
            sys.exit()
        elif o == "-r":
            ipc_local_rid = atoid(a)
        elif o == "-q":
            ipc_net_rid = atoid(a)
        elif o == "-t":
            debug_level = int(a)
        elif o == "-c":
            debug_colors = True
        else:
            assert False, "Unhandled option: %r" % o

    def signal_handler(signal_number, stack_frame):
        raise KeyboardInterrupt()
    signal.signal(signal.SIGINT, signal_handler)
    
    debug_init(debug_level, debug_colors)
    
    kwargs = {"sid_pattern": sid_pattern,
              "ipc_sid": ipc_sid,
              "ipc_local_rid": ipc_local_rid,
              "ipc_net_rid": ipc_net_rid}
    
    s = Subscriber(**kwargs)
    s.laird_listen()


if __name__ == "__main__":
    main()

