/*
* Copyright (C) 2009, Oy L M Ericsson Ab, NomadicLab <pubsub@nomadiclab.com>
*
* This program is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License version 2 as
* published by the Free Software Foundation.
*
* Alternatively, this software may be distributed under the terms of the BSD
* license.
*
* See LICENSE and COPYING for more details.
*/

/* psirpd_net.h */
#include <sys/types.h>
#include <string.h>
#include <stdlib.h>

#include <sys/socket.h>
#include <net/if_dl.h>
#include <net/if.h>

#include <sys/param.h>
#include <libpsirp.h>
#include "../../libpsirp/src/psirp_debug.h"

#include "psirp_common_types.h"
#include "psirp.h"
#include "psirpd_hdrs.h"
#include "psirpd_net.h"
#include "psirpd_out_q.h"
#include "psirpd_packet.h"

LIST_HEAD(out_q_list, out_q_list_item) out_q_list_head;

static out_q_list_item_t *item_out = NULL;

psirp_error_t
psirpd_out_q_init(void) {

    /* XXX should this be called from somewhere?? */

    LIST_INIT(&out_q_list_head); 
    return PSIRP_OK;
}

void
psirpd_out_q_cleanup(void) {

    /* XXX should this be called from somewhere?? */

    out_q_list_item_t *p1, *p2;

	LIST_FOREACH_SAFE(p1, &out_q_list_head, entries, p2) {
        LIST_REMOVE(p1, entries);
        PSIRP_FREE(p1);
    }
}

/** 
 * Function that is called for allocating an out queue item.
 * @return out queue list item
 */
out_q_list_item_t*
psirpd_out_q_create_item(psirp_id_t *sidp,
                         psirp_id_t *ridp,
                         psirp_id_t *vridp,
			 psirp_fid_t *fidp,
			 psirp_pub_t *pubp,
			 u_int32_t *countp,
			 u_int64_t seqnum) {
    //psirp_error_t err;
    out_q_list_item_t *item;

    ENTER();
    
    PSIRP_MALLOC(item, sizeof(out_q_list_item_t));
    memset(item, 0, sizeof(out_q_list_item_t));

#if 0
    item->fid = *fidp;
    item->rid = *ridp;
    item->sid = *sidp;
#else
    memcpy(&item->fid, fidp, sizeof(psirp_fid_t));
    memcpy(&item->sid, sidp, sizeof(psirp_id_t));
    memcpy(&item->rid, ridp, sizeof(psirp_id_t));
    memcpy(&item->vrid, vridp, sizeof(psirp_id_t));
#endif
    
    item->pubp   = pubp;
    item->countp = countp;
    item->seqnum = seqnum;
    
    RETURN item;
}

/** 
 * Function that is called for freeing memory that is allocated for
 * the out queue item.
 * @param item pointer to the out queue item
 * @return void
 */
void
psirpd_out_q_free_item(out_q_list_item_t *item) {

    ENTER();
    
    if ((int32_t)*(item->countp) < 0) {
	
	PSIRP_DEBUG(PSIRP_DBG_INFO,
		    "All chunks sent");
	
	psirp_free(*item->pubp);
	*item->pubp = NULL;
	
	PSIRP_FREE(item->pubp);
	item->pubp = NULL;
	
	PSIRP_FREE(item->countp);
	item->countp = NULL;
    }
    
    PSIRP_FREE(item);
    
    RETURN;
}

/** 
 * Function that is called for adding the out queue item to the 
 * out queue list.
 * @param item1 pointer to the out queue item
 * @return void
 */
void
psirpd_out_q_add_item(out_q_list_item_t *item1) {

    ENTER();
    
    if (LIST_EMPTY(&out_q_list_head)) {
        item_out = item1;
    }

    LIST_INSERT_HEAD(&out_q_list_head, item1, entries);

    RETURN;
}
              
/** 
 * Function is called during every select() timeout. The function loops
 * the out queue list and sends a single packet per timeout.  
 * @return On sucess: PSIRP_OK <br>On failure: PSIRP_FAIL_ error code
 */
psirp_error_t
psirpd_out_q_send(void) {
    
    u_int8_t *addr;
    out_q_list_item_t *item_next = NULL;
    psirp_error_t retval = PSIRP_OK;
    int i;

    ENTER();

    item_out = LIST_FIRST(&out_q_list_head);

    for (i = 0; i < PSIRP_OUTQ_BATCH_SIZE; i++) {
	if (NULL == item_out) {
	  retval = PSIRP_OK;
	  break;
	}
	
	addr = (u_int8_t *)(psirp_pub_data(*(item_out->pubp))
			    + item_out->seqnum*PSIRP_CHUNK_SIZE);
	
	PSIRP_DEBUG(PSIRP_DBG_GARB,
		    "Send chunk out (count=%l, seqnum=%lu, addr=%p)",
		    *(item_out->countp), item_out->seqnum, addr);
	retval = psirpd_packet_out_data(addr, 
					PSIRP_CHUNK_SIZE, 
					item_out);
	*(item_out->countp) -= 1;
	
	item_next = LIST_NEXT(item_out, entries);
	if (NULL == item_next) {
	    PSIRP_DEBUG(PSIRP_DBG_GARB, "We are at the end of the list");
	}
	
	LIST_REMOVE(item_out, entries);
	psirpd_out_q_free_item(item_out);
	
	item_out = item_next;
    }
    
    PSIRP_DEBUG(PSIRP_DBG_GARB, "Sent out %d packets at once", i);

    RETURN retval;
}

struct timespec *
psirpd_out_q_next_timeout(struct timespec *time) {

    ENTER();

    if (NULL != item_out) {
        /* Things in outqueue. */
        memset(time, 0, sizeof(time));
        time->tv_nsec = 50000000; /* 1 ms = 1 000 000 ns */
        time->tv_sec = 0;
        RETURN time;
    } else {
        /* Nothing in outqueue. */
        RETURN NULL;
    }
}
