#
# Copyright (C) 2009, Oy L M Ericsson Ab, NomadicLab <pubsub@nomadiclab.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2 as
# published by the Free Software Foundation.
#
# Alternatively, this software may be distributed under the terms of the BSD
# license.
#
# See LICENSE and COPYING for more details.
#


# Version check

import sys
assert sys.version_info[0] == 2 and sys.version_info[1] >= 6, \
    "Python 2.6 or a newer 2.x version required."
del sys


from psirp import libpsirp_py
import errno, select, time #, signal


# Errors

class PubSubError(Exception):
    """Parent class for pub/sub errors."""
    pass


class PubStateError(PubSubError):
    """Parent class for pub/sub state errors."""
    pass

class NotInitializedError(PubStateError):
    """Publication object not initialized."""
    pass

class NotPublishedError(PubStateError):
    """Publication not published."""
    pass


class PubSubNotImplementedError(PubSubError, NotImplementedError):
    """A feature is not implemented, is unsupported, or has been disabled."""
    pass

class NoEventHandlerError(PubSubNotImplementedError):
    """Publication has no event handler"""
    pass


class PubSubIOError(PubSubError, EnvironmentError):
    """Parent class for pub/sub I/O errors."""
    _errno = 0
    
    def __init__(self, err=0, str=""):
        EnvironmentError.__init__(self, err, str)

class NotFoundError(PubSubIOError):
    """Parent class for "not found" errors."""
    pass

class ScopeNotFoundError(NotFoundError):
    """Scope not found."""
    _errno = errno.ESRCH
    
    def __init__(self, str=""):
        NotFoundError.__init__(self, self._errno, str)

class PubNotFoundError(NotFoundError):
    """Publication not found in scope."""
    _errno = errno.ENOENT
    
    def __init__(self, str=""):
        NotFoundError.__init__(self, self._errno, str)


#class PubSubInterrupt(PubSubError, KeyboardInterrupt):
#    """Pub/Sub operation interrupted."""
#    pass
#
#def _interrupt_handler(signal_number, stack_frame):
#    raise PubSubInterrupt()
#signal.signal(signal.SIGINT, _interrupt_handler)


# Constants

from libpsirp_py import \
    PSIRP_ID_LEN, PSIRP_FID_LEN, \
    NOTE_PUBLISH, NOTE_SUBSCRIBE, NOTE_UNMAP, \
    PS_PUB_SCOPE, PS_PUB_DATA, PS_PUB_VERSION, PS_PUB_PAGE


# Internal functions

def _create(length):
    v = libpsirp_py.psirp_create(length) # -> (retval, pupb)
    
    if v is None or len(v) < 2 or v[0] != 0 or v[1] is None:
        err = (v[0] if v is not None and len(v) >= 1 else 0)
        raise PubSubIOError(err, "Create failed%s"
                            % (" (kernel module probably not loaded)"
                               if err == -1 else ""))
    
    return v[1] # pupb

def _subscribe(sync, *args):
    v = (libpsirp_py.psirp_py_subscribe_sync_(*args) if sync
         else libpsirp_py.psirp_subscribe(*args)) # -> (retval, pupb)
    
    if v is None or len(v) < 2 or v[0] != 0 or v[1] is None:
        err = (v[0] if v is not None and len(v) >= 1 else 0)
        if err == PubNotFoundError._errno:
            raise PubNotFoundError("Publication not found")
        elif err == ScopeNotFoundError._errno:
            raise ScopeNotFoundError("Scope not found")
        else:
            raise PubSubIOError(err, "Subscribe failed%s"
                                % (" (kernel module probably not loaded)"
                                   if err == -1 else ""))
        
    return v[1] # pupb

def _copybuf(buf):
    return buffer(buf[:])


# Public functions for create, subscribe, publish, and Id conversions

def create(length):
    """Creates a new publication."""
    return Publication._init(_create(length)) # -> Publication
    
def subscribe(sid, rid):
    """Subscribes to a publication."""
    return Publication._init(
        _subscribe(False, sid, rid),
        True) # -> Publication
    
def subscribe_(sidstr, ridstr):
    """Subscribes to a publication. Ids are given as hex strings."""
    sid = atoid(sidstr)
    rid = atoid(ridstr)
    return subscribe(sid, rid) # -> Publication

def wait_for_rid(sid, rid, t0=None, timeout=None, pskq=None,
                 exception=PubNotFoundError):
    """Wait until rid appears in scope sid, or until timeout."""

    if t0 is None and timeout is not None:
        t0 = time.time()
    to = None
    
    if pskq is None:
        pskq = PubSubKQueue()
    
    scope = subscribe(sid, sid) # Assume that this scope exists
    rids = scope.get_rids(1)    # RId 0 is the SId
    rid_count = len(rids)
    if rid in rids:
        # RId found in scope SId
        return
    del rids
    
    pskq.register(scope, True, True) # Note that scope gets unmapped here
    
    try:
        while True:
            if timeout is not None:
                t1 = time.time()
                td = t1 - t0
                if td > timeout:
                    raise exception("Timeout")
                to = timeout - td
                
            publ = None            
            try:
                publ = pskq.listen(1, to)
            except Exception, e:
                raise exception(repr(e))
                    
            # Error checks
            if publ is None:
                raise exception("Publication list is None")
            elif len(publ) == 0:
                raise exception("Timeout")
            elif publ[0][1] is not scope:
                raise exception("Unknown publication returned")
            
            scope = publ[0][1]
            new_rids = scope.get_rids(rid_count)
            if rid in new_rids:
                # RId found in scope SId
                return
            rid_count += len(new_rids)
    finally:
        pskq.unregister(scope)

def subscribe_sync(sid, rid, timeout=None):
    """Synchronous subscribe. (Currently only works with "real" SIds.)"""
    
#    return Publication._init(
#        _subscribe(True, sid, rid, (timeout if timeout is not None else -1)),
#        True) # -> Publication
    
    t0 = time.time()
    pskq = PubSubKQueue()
    sid0 = buffer(PSIRP_ID_LEN*'\x00')
    
    try:
        try:
            return subscribe(sid, rid)
        
        except ScopeNotFoundError:
            # The scope doesn't yet exist locally.
            # Subscribe to Scope 0 first, and
            # see if the new scope appears.
            wait_for_rid(sid0, sid, t0, timeout, pskq,
                         ScopeNotFoundError)
            
            # Sub-scope found in given scope;
            # try to subscribe to the publication;
            # goes to outer except if fails.
            return subscribe(sid, rid)
        
    except PubNotFoundError:
        # The scope seems to exist locally, but the publication is not in it.
        # Subscribe to the scope, and see if the publication appears.
        
        wait_for_rid(sid, rid, t0, timeout, pskq,
                     PubNotFoundError)
        return subscribe(sid, rid)
    
    finally:
        pskq.close()

def subscribe_sync_(sidstr, ridstr, timeout=None):
    """Synchronous subscribe. Ids are given as hex strings."""
    sid = atoid(sidstr)
    rid = atoid(ridstr)
    return subscribe_sync(sid, rid, timeout) # -> Publication

def publish(sid, rid, pub):
    """Publishes a publication."""
    pub.publish(sid, rid)

def publish_(sidstr, ridstr):
    """Publishes a publication. Ids are given as hex strings."""
    pub.publish(sidstr, ridstr)

def atoid(a):
    """Hex string to binary identifier conversion."""
    return buffer(libpsirp_py.psirp_py_atoid(a))

def idtoa(id):
    """Binary identifier to hex string conversion."""
    return libpsirp_py.psirp_idtoa(id)

def idstoa(sid, rid):
    """Binary identifier to hex string conversion."""
    return libpsirp_py.psirp_idstoa(sid, rid)


# Publication objects

class Publication(object):
    
    """Publication object.
    
    Important notice:
    
    If references to binary data (RIds, buffers, etc.) are stored in
    variables, the underlying data may get unmapped from the
    application's memory space when the publication object is garbage
    collected, or when a new version is acquired via the kevent
    system. This results in a segmentation fault when the old data is
    accessed. Therefore, any data to be stored for later use must
    either be copied, or then the programmer must ensure that the
    publication object and the specific version are kept in memory for
    as long as the data needs to be accessible.
    """
    
    def __init__(self, pub=None):
        """Should not be called explicitly from external sources."""
        
        if pub is not None:
            raise DeprecationWarning()
        
        self.__pubp = None
        self.__published = False
        self.__free_on_del = False
    
    def __del__(self):
        """Destructor that calls psirp_free() by default."""
        
        if self.__pubp is not None and self.__free_on_del:
            libpsirp_py.psirp_free(self.__pubp)
            self.__pubp = None
            self.__free_on_del = False
    
    @classmethod
    def _init(cls, pubp, published=False, free_on_del=True):
        pub_type = libpsirp_py.psirp_py_type(pubp)
        if pub_type in _PUB_CLASSES:
            pub_cls = _PUB_CLASSES[pub_type]
        else:
            raise PubSubError("Bad publication type (%d)" % pub_type)
        
        publication = pub_cls() # -> Publication
        publication.__pubp = pubp
        publication.__published = published
        publication.__free_on_del = free_on_del
        return publication
    
    @staticmethod
    def _publish(sid, rid, pubp):
        v = libpsirp_py.psirp_publish(sid, rid, pubp);  # -> retval
        
        if v != 0:
            raise PubSubIOError(v, "Publish failed")

    def _checkinit(self):
        if self.__pubp is None:
            raise NotInitializedError()
    
    def _checkpub(self):
        if not self.__published:
            raise NotPublishedError("Publication not yet published")
    
    def _checkinitpub(self):
        if self.__pubp is None:
            raise NotInitializedError()
        if not self.__published:
            raise NotPublishedError("Publication not yet published")
    
    @classmethod
    def from_pubp(cls, pubp, published=False, free_on_del=False):
        """Form a Publication object from a pointer."""
        return cls._init(pubp, published, free_on_del)

    def resubscribe(self):
        """Re-subscribes to the same SId and RId and returns a new object.
        
        Returns the latest version."""
        # Design choice: we could also replace self.__pubp, but here we
        #                return a new object.
        return subscribe(self.sid, self.rid)
    
    def publish(self, sid=None, rid=None):
        """Publishes the publication."""
        
        self._checkinit()
        
        if sid is None:
            sid = self.sid
        if rid is None:
            rid = self.rid
        
        self._publish(sid, rid, self.__pubp)
        self.__published = True # XXX: ?
    
    def republish(self):
        """Re-publishes the publication with the same SId and RId."""
        self.publish(None, None)
    
    def publish_(self, sidstr, ridstr):
        """Publishes the publication. Ids are given as hex strings."""
        
        sid = atoid(sidstr)
        rid = atoid(ridstr)
        return self.publish(sid, rid)
    
    @property
    def pubp(self):
        """Pointer to the underlying psirp_pub_t structure."""
        self._checkinit()
        return self.__pubp
    
    @property
    def datap(self):
        """Pointer to the underlying data."""
        self._checkinit()
        return libpsirp_py.psirp_pub_data(self.__pubp)
    
    @property
    def buffer(self):
        """Get publication data as a read-write buffer."""
        # XXX: A new buffer object is created every time this member
        #      is accessed, which could lead to inefficient code. On
        #      the other hand, replacing it with, e.g., a get_buffer()
        #      function would easily lead to either ugly code, or code
        #      that causes a segmentation fault when the publication
        #      has been freed but the buffer object still exists.
        self._checkinit()
        return libpsirp_py.psirp_py_buffer(self.__pubp)
        
    @property
    def len(self):
        """Publication length."""
        self._checkinit()
        return libpsirp_py.psirp_pub_data_len(self.__pubp)
    
    @property
    def sid(self):
        """Scope identifier (SId)."""
        self._checkinitpub()
        return libpsirp_py.psirp_pub_sid(self.__pubp)
    
    @property
    def rid(self):
        """Rendezvous identifier (RId)."""
        self._checkinitpub()
        return libpsirp_py.psirp_pub_rid(self.__pubp)
    
    @property
    def vrid(self):
        """Current version-RId."""
        self._checkinitpub()      
        return libpsirp_py.psirp_pub_current_version(self.__pubp)

    @property
    def version_index(self):
        """Index of current version."""
        self._checkinitpub()
        return libpsirp_py.psirp_pub_current_version_index(self.__pubp)

    @property
    def version_count(self):
        """Total number of versions."""
        self._checkinitpub()
        return libpsirp_py.psirp_pub_version_count(self.__pubp)
    
    def get_vrids(self, index=0):
        """List of version-RIds."""
        
        self._checkinitpub()
        vrids = libpsirp_py.psirp_py_pub_get_vrids(self.__pubp, index)
        if vrids is None:
            raise PubSubError()
        return vrids
    
    @property
    def fd(self):
        """File descriptor for kevents."""
        self._checkinitpub()
        return libpsirp_py.psirp_pub_fd(self.__pubp)

    @property # XXX: Make settable?
    def type(self):
        """Publication type."""
        self._checkinit()
        return libpsirp_py.psirp_py_type(self.__pubp)

    @property
    def addr(self):
        """Publication memory address for kevents."""
        self._checkinit() # XXX: initpub?
        # XXXXXX: Return a 32-bit address that can be used as udata in
        #         Python's kevent implementation.
        return int(int(self.__pubp) & ~0x800000000)
    
    @property
    def ridstr(self):
        """RId as a hexadecimal string."""
        return libpsirp_py.psirp_idtoa(self.rid)

    @property
    def sidstr(self):
        """SId as a hexadecimal string."""
        return libpsirp_py.psirp_idtoa(self.sid)

    @property
    def vridstr(self):
        """Version-RId as a hexadecimal string."""
        return libpsirp_py.psirp_idtoa(self.vrid)
    
    def handle_event(self, event, pub):
        """Event handler function. Abstract by default."""
        raise NoEventHandlerError()
    
    def __str__(self):
        if not self.__published:
            return repr(self)
        return "<%s SId=%s RId=%s vRId=%s len=%d fd=%d>" \
            % (self.__class__.__name__,
               ("" if not isinstance(self, DataPublication) else self.sidstr),
               self.ridstr,
               self.vridstr,
               self.len,
               self.fd)

class DataPublication(Publication):
    pass

class ScopePublication(DataPublication):
    def __init__(self):
        """Should not be called explicitly from external sources."""
        super(type(self), self).__init__()
    
    # New methods and members
    
    @property
    def rid_count(self):
        """Number of RIds in scope."""
        self._checkinitpub()
        return libpsirp_py.psirp_scope_rid_count(self.pubp)
    
    def get_rids(self, index=0):
        """List of RIds in scope."""
        
        self._checkinitpub()
        rids = libpsirp_py.psirp_py_scope_get_rids(self.pubp, index)
        if rids is None:
            raise PubSubError()
        return rids
    
    # Disabled methods and members
    
    def publish(self, sid=None, rid=None):
        """Disabled."""
        raise PubSubNotImplementedError(
            "Scope publication cannot be published")
    # XXX: publish_ and republish?
    
    @property
    def buffer(self):
        """Disabled."""
        raise PubSubNotImplementedError(
            "Raw scope data cannot be accessed")

class VersionPublication(Publication):
    def __init__(self):
        """Should not be called explicitly from external sources."""
        super(type(self), self).__init__()

    # New methods and members
    
    def get_prids(self, index=0):
        """List of page-RIds."""
        
        self._checkinitpub()
        prids = libpsirp_py.psirp_py_version_get_prids(self.pubp, index)
        if prids is None:
            raise PubSubError()
        return prids
    
    # Modified methods and members
    
    @property
    def rid(self):
        """RId of the data publication."""
        return super(type(self), self).sid
    
    @property
    def vrid(self):
        """Version-RId of this version."""
        return super(type(self), self).rid
    
    # Disabled methods and members

    def publish(self, sid=None, rid=None):
        """Disabled."""
        raise PubSubNotImplementedError(
            "Re-publishing a version publication is not supported") # XXX
    
    @property
    def sid(self):
        """Disabled."""
        raise PubSubNotImplementedError(
            "Version publication does not have a SId")
    
    @property
    def version_index(self):
        """Disabled."""
        raise PubSubNotImplementedError(
            "Version publication does not have versions")

    @property
    def version_count(self):
        """Disabled."""
        raise PubSubNotImplementedError(
            "Version publication does not have versions")
    
    @property
    def vrids(self):
        """Disabled."""
        raise PubSubNotImplementedError(
            "Version publication does not have versions")

class PagePublication(Publication):
    def __init__(self):
        """Not supported."""
        raise  PubSubNotImplementedError(
            "Page publications are not supported") # XXX
    

_PUB_CLASSES = {PS_PUB_DATA:    DataPublication,
                PS_PUB_SCOPE:   ScopePublication,
                PS_PUB_VERSION: VersionPublication,
                PS_PUB_PAGE:    PagePublication}


# Pub/sub kqueue

class PubSubKQueue(object):
    def __init__(self, kq=None):
        self.kq = select.kqueue() if kq is None else kq
        self.pubs = {}
    
    def close(self):
        """Close the kqueue."""
        
        self.kq.close()
        self.pubs.clear()
    
    def register(self, pub, update=False, unmap=False):
        """Register to kevents for a publication.
        
        Args:
        pub    - publication object
        update - True:  update pub to the latest version when an event occurs
                 False: keep the current version and leave pub untouched
        unmap  - True:  unmap the old version from our memory space
                 False: keep the old version mapped
        """
        
        ident = pub.fd
        filter = select.KQ_FILTER_VNODE
        flags = select.KQ_EV_ADD | select.KQ_EV_CLEAR
        fflags = libpsirp_py.NOTE_PUBLISH
        if unmap:
            fflags |= libpsirp_py.NOTE_UNMAP # XXX: free_on_del?
        data = 0x0
        if update:
            udata = pub.addr
        else:
            udata = 0x0
        
        evf = select.kevent(ident, filter, flags, fflags, data, udata)
        self.kq.control([evf], 0, None)
        
        self.pubs[ident] = pub
    
    def custom_command(self, ident,
                       filter=select.KQ_FILTER_VNODE,
                       flags=select.KQ_EV_ADD|select.KQ_EV_CLEAR,
                       fflags=0x0,
                       data=0x0,
                       udata=0x0):
        """Register or unregister any kevents on this kqueue."""
        
        evf = select.kevent(ident, filter, flags, fflags, data, udata)
        self.kq.control([evf], 0, None)
    
    def unregister(self, pub):
        """Unregister a publication."""
        
        ident = pub.fd
        filter = select.KQ_FILTER_VNODE
        flags = select.KQ_EV_DELETE
        
        evf = select.kevent(ident, filter, flags)
        self.kq.control([evf], 0, None)
        
        del self.pubs[ident]
    
    def listen(self, n=1, timeout=None):
        """Listen to kevents.
        
        Args:
        n       - max. number of kevents to return in a list
        timeout - kevent timeout value
        """
        
        publ = None
        
        try:
            evl = self.kq.control(None, n, timeout) # listen to kevents
            
            if evl is not None:
                publ = []
                for ev in evl:
                    if ev.ident in self.pubs:
                        publ.append((ev, self.pubs[ev.ident]))
                    else:
                        publ.append((ev, None))
                return publ
                
        except (KeyboardInterrupt, SystemExit):
            raise


def subscribe_versions(pub, start_index, stop_index=None):
    """Subscribes to the specified range of versions.
    
    Args:
    pub         - publication object
    start_index - index of first version to include
    stop_index  - index of last version to include (default: last index)
    
    Usage scenario:
    If a publication is updated multiple times during the time when a kevent
    is being processed, only one kevent is returned, and the publication data
    is set to point to the latest version. With this function, the "missed"
    intermediate versions can be acquired conveniently if they're needed.
    
    Example:
    subscribe_versions(pub, previous_index+1, pub.version_index-1)
    """
    # XXX: Publication objects could implement similar methods as well...
    
    vcount = pub.version_count
    if stop_index is None:
        stop_index = vcount-1
    if stop_index >= vcount or start_index < 0 or start_index > stop_index:
        raise IndexError("Version indexes %d..%d out of range"
                         % (start_index, stop_index))
    
    rid = pub.rid
    vrids = pub.get_vrids(start_index)
    n = (stop_index - start_index) + 1
    
    vpubs = []
    for i in xrange(n):
        vpubs.append(subscribe(rid, vrids[i]))
    
    return vpubs
