/*
* Copyright (C) 2009, Oy L M Ericsson Ab, NomadicLab <pubsub@nomadiclab.com>
*
* This program is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License version 2 as
* published by the Free Software Foundation.
*
* Alternatively, this software may be distributed under the terms of the BSD
* license.
*
* See LICENSE and COPYING for more details.
*/

#ifndef _PS_H
#define _PS_H

/* User space applications
 * 
 * Required definitions:
 * - kevent flags
 * - ps_pub (user space view) && accessors
 *
 * User space libraries
 *
 * Required definitions:
 * - kevent flags
 * - ps_pub (almost full kernel view [includes meta data])
 * - syscalls && args
 * - 
 *
 * Kernel
 *
 * Required defintions:
 * - Everything but...
 *   - user space (re)definitions
 *
 */

/*
 * The system call module name.  Shared between user level and library.
 *
 * If you change this, be sure to change SYSCALL_MODULE definition in
 * the end of ps_syscall.c
 */

#define PS_SYSCALL_MODULE       ps_syscall
#define PS_SYSCALL_MODULE_NAME "ps_syscall"

/*
 * UID and GID for versions and pages, which are shared between publications
 * therefore not really "owned" by anybody
 */
#ifndef UID_NOBODY
#define UID_NOBODY 65534
#endif
#ifndef GID_NOBODY
#define GID_NOBODY 65534
#endif


/*
 * New VNODE kevent types.  NOTE_PUBLISH returns publish events, either on
 * a publication vnode, in which case udata points to ps_pub_t, or
 * on the root vnode, in which case udata points to ps_scope_knote.
 * NOTE_SUBSCRIBE is less well defined at this point, but can be used
 * at least on the root vnode, using ps_scope_knote.
 * 
 * NOTE_UNMAP is not really a traditional kevent, but a shorthand to
 * unmap previous publication version at kevent(2) call time.
 */
#define NOTE_PUBLISH    0x0100
#define NOTE_SUBSCRIBE  0x0200
#define NOTE_UNMAP      0x0400

/* Identifiers */

#define PSIRP_ID_LEN           32
#define PSIRP_FID_LEN          PSIRP_ID_LEN

struct psirp_id {
    u_int8_t id[PSIRP_ID_LEN]; /**< Common ID structure (P:L, Scope, etc) */
};

typedef struct psirp_id psirp_id_t, psirp_fid_t;

/* XXX: Should use different magic numbers instead of the enum */
enum ps_pub_type { PS_PUB_UNINITIALISED=0, PS_PUB_UNKNOWN, 
		   PS_PUB_SCOPE, PS_PUB_DATA, PS_PUB_VERSION, PS_PUB_PAGE,
                   PS_PUB_TYPE_COUNT };
#define PS_PUB_TYPE(meta)       ((meta)->pm_type)
#define PS_PUB_MUTABLE(meta)    ((meta)->pm_type < PS_PUB_VERSION)
#define PS_PUB_CHECK_TYPE(meta) (PS_PUB_TYPE(meta) > PS_PUB_UNINITIALISED && PS_PUB_TYPE(meta) < PS_PUB_TYPE_COUNT)

struct ps_meta_hdr {
    psirp_id_t        	     _pm_h_magic;     /* Metadata type */
    psirp_id_t               _pm_h_id;	      /* RID for this publication/version/page */
    enum ps_pub_type         _pm_h_type;      /* XXX: To be merger with pm_magic? */
    off_t                    _pm_h_size;      /* Bytes in the data vm_object */
    union {
	struct {
	    int              _pm_m_vers_count;
	    struct mtx      *_pm_m_interlock;
	} _pm_u_mutable;
	struct {
	    int              _pm_s_page_count; /* Fixed at creation time */
	} _pm_u_static;
    } _pm_h_u;
};
#define PS_META_SUB_OBJECT_COUNT ((PAGE_SIZE - sizeof(struct ps_meta_hdr)) / sizeof(psirp_id_t))
struct ps_meta {
    struct ps_meta_hdr       _pm_h;
    psirp_id_t               pm_sub_object[PS_META_SUB_OBJECT_COUNT];
};
typedef struct ps_meta *ps_meta_t;
#define pm_magic      _pm_h._pm_h_magic
#define pm_type       _pm_h._pm_h_type
#define pm_id         _pm_h._pm_h_id
#define pm_size       _pm_h._pm_h_size
#define pm_interlock  _pm_h._pm_h_u._pm_u_mutable._pm_m_interlock
#define pm_vers_count _pm_h._pm_h_u._pm_u_mutable._pm_m_vers_count
#define pm_page_count _pm_h._pm_h_u._pm_u_static._pm_s_page_count

struct psfs_pub;

#if defined(_LIBPSIRP) || defined(_KERNEL)
/*
 * Library/kernel interface.  Opaque to the user apps.
 */
struct psfs_pub {
    ps_meta_t    pub_meta;		/* Pointer to the local metadata */
    u_int64_t    pub_mlen;		/* Length of the metadata == PSFS_MD_SIZE */
    caddr_t      pub_data;		/* Pointer to the actual data */
    u_int64_t    pub_dlen;		/* Length of the actual data */
    int          pub_fd;                /* Dir fd, maintained by the library */
    int          pub_vidx;              /* Version index */
    psirp_id_t   pub_sid;               /* SID used when subscribed/published*/
  
};

typedef enum ps_vmo { VMO_META, VMO_DATA, VMO_NUM } ps_vmo_t;

#endif

typedef struct psfs_pub psfs_pub_t, *psirp_pub_t;

#define PSFS_PUB_META(p)     ((p)->pub_meta)
#define PSFS_PUB_META_LEN(p) ((p)->pub_mlen)
#define PSFS_PUB_DATA(p)     ((p)->pub_data)
#define PSFS_PUB_DATA_LEN(p) ((p)->pub_dlen)
#define PSFS_PUB_FD(p)       ((p)->pub_fd)

#endif /* _PS_H */
